(* Type représentant un trie.
   Pour ce type et la fonction find, voir:

    https://gist.github.com/Tchou/c0c305621bc909e189db90821d78eb6f
*)

type 'a trie =
    Node of ('a option * (char * 'a trie) list)
  (** [empty] est le trie vide. *)
let empty = Node (None, [])

(** [is_empty t] est vrai si et seulement si [t] est vide. *)
let is_empty t =
  match t with
    Node (None, []) -> true
  | _ -> false

(** [find_trie key t] renvoie le (sous)-trie accessible
    dans [t] en suivant la clé [key].
    Lève l'exception [Not_found] si [key] n'est pas dans [t] *)
let find_trie key t =
  let rec find_node i t =
    if i = String.length key then t
    else
      match t with Node (_, l) -> find_list i l

  and find_list i l =
    let ci = key.[i] in
    match l with
      [] -> raise Not_found
    | (d, t) :: ll ->
      if ci > d then find_list i ll
      else if ci = d then find_node (i+1) t
      else raise Not_found
  in
  find_node 0 t

(** [find key t] renvoie la valeur accessible dans
    [t] en suivant la clé [key]. Lève l'exception [Not_found]
    si [key] n'est pas dans [t]. *)
let find key t =
  match find_trie key t with
    Node (Some v, _) -> v
  | Node (None, _) -> raise Not_found

(** Q1.1 On suit la même structure que find. *)
let add key v t =
  let rec insert_node i t =
    match t with
    | Node (o, l) ->
      if i = String.length key then Node (Some v, l)
      else Node (o, insert_list i l)
  and insert_list i l =
    let ci = key.[i] in
    match l with
    | [] -> [ (ci, insert_node (i + 1) empty) ]
    | (d, t) :: ll ->
      if ci > d then (d, t) :: insert_list i ll
      else if ci = d then (ci, insert_node (i + 1) t) :: ll
      else (ci, insert_node (i + 1) empty) :: l (* attention, ici l pour laisser (d,t)*)
  in
  insert_node 0 t
(* Q1.2 Test : *)
(* On essaye d'écrire un test un peu générique *)
let add_from_list l t =
  List.fold_left (fun acc (k, v) -> add k v acc) t l
let rev_implode l =
  String.concat "" (List.rev_map (String.make 1) l)
